# coding=utf-8

import logging
import asyncio
import signal
import json
from asyncio import CancelledError
import os

import aiohttp
import async_timeout

log = logging.getLogger(__name__)

DEFAULT_UPDATE_INTERVAL = 60
DEFAULT_FREEHP_ADDRESS = 'localhost:6256'
DEFAULT_SQUID = 'squid3'
DEFAULT_TIMEOUT = 30
DEFAULT_MAX_NUM = 0

PEER_CONF = "cache_peer {} parent {} 0 no-query weighted-round-robin weight=1 connect-fail-limit=2 allow-miss max-conn=5 name={}\n"


class Squid:
    def __init__(self, dest_file, tpl_file, freehp_address=DEFAULT_FREEHP_ADDRESS, squid=DEFAULT_SQUID,
                 max_num=DEFAULT_MAX_NUM, update_interval=DEFAULT_UPDATE_INTERVAL, timeout=DEFAULT_TIMEOUT, once=False):
        self.loop = asyncio.new_event_loop()
        self._dest_file = dest_file
        with open(tpl_file, 'rb') as f:
            self._template = f.read().decode()
        self._squid = squid
        self._max_num = max_num
        self._update_interval = update_interval
        self._timeout = timeout
        self._once = once
        if freehp_address.startswith('http://') or freehp_address.startswith('https://'):
            self._freehp_address = freehp_address
        else:
            self._freehp_address = 'http://' + freehp_address
        if not self._freehp_address.endswith('/'):
            self._freehp_address += '/'
        self._freehp_address += 'proxies'
        if max_num > 0:
            self._freehp_address = '{}?count={}'.format(self._freehp_address, max_num)
        self._future = None
        self._is_running = False

    def start(self):
        if not self._is_running:
            self._is_running = True
            asyncio.set_event_loop(self.loop)
            if self._once:
                log.info('Run only once')
                self.loop.run_until_complete(self._maintain_squid())
                log.info('Task is done')
            else:
                self._future = asyncio.ensure_future(self._maintain_squid_task(), loop=self.loop)
                self.loop.add_signal_handler(signal.SIGINT,
                                             lambda loop=self.loop: asyncio.ensure_future(self.shutdown(), loop=loop))
                self.loop.add_signal_handler(signal.SIGTERM,
                                             lambda loop=self.loop: asyncio.ensure_future(self.shutdown(), loop=loop))
                log.info('Run forever')
                try:
                    self.loop.run_forever()
                except Exception:
                    log.error("Unexpected error occurred when run loop", exc_info=True)
                    raise

    async def shutdown(self):
        if not self._is_running:
            return
        self._is_running = False
        log.info("Shutdown now")
        if self._future:
            self._future.cancel()
        self._future = None
        await asyncio.sleep(0.001, loop=self.loop)
        self.loop.stop()

    async def _maintain_squid_task(self):
        while True:
            await self._maintain_squid()
            await asyncio.sleep(self._update_interval, loop=self.loop)

    async def _maintain_squid(self):
        data = []
        try:
            async with aiohttp.ClientSession(loop=self.loop) as session:
                with async_timeout.timeout(self._timeout, loop=self.loop):
                    async with session.request("GET", self._freehp_address) as resp:
                        body = await resp.read()
                        data = json.loads(body.decode('utf-8'))
                        log.debug('Get %s proxies', len(data))
        except CancelledError:
            raise
        except Exception:
            log.error("Error occurred when get proxies from '%s'", self._freehp_address, exc_info=True)

        if len(data) > 0:
            try:
                self._reconfigure_squid(data)
            except Exception:
                log.error("Error occurred when reconfigure squid", exc_info=True)

    def _reconfigure_squid(self, proxies):
        lines = [self._template, '\n# cache_peer configuration\n']
        for p in proxies:
            host, port = p.split(':')
            lines.append(PEER_CONF.format(host, port, host + '.' + port))
        with open(self._dest_file, 'w') as f:
            f.writelines(lines)
        log.info('Reconfigure squid with %s proxies, conf=%s', len(proxies), self._dest_file)
        try:
            if os.system('{} -k reconfigure'.format(self._squid)) != 0:
                raise RuntimeError
        except RuntimeError:
            with open(self._dest_file, 'w') as f:
                f.write(self._template)
            raise



# coding=utf-8

import json
import random
import asyncio
import logging
from asyncio import CancelledError

import aiohttp
import async_timeout

log = logging.getLogger(__name__)


class HttpbinChecker:
    def __init__(self, *, loop=None, timeout=10):
        self._loop = loop or asyncio.get_event_loop()
        self._timeout = int(timeout)

    @classmethod
    def from_manager(cls, manager):
        c = manager.config.get("httpbin_checker") or {}
        return cls(loop=manager.loop, **c)

    async def check_proxy(self, addr):
        if not addr.startswith("http://"):
            proxy = "http://{0}".format(addr)
        else:
            proxy = addr
        try:
            async with aiohttp.ClientSession(loop=self._loop) as session:
                with async_timeout.timeout(self._timeout, loop=self._loop):
                    seed = str(random.randint(0, 99999999))
                    url = "http://httpbin.org/get?seed={}".format(seed)
                    async with session.request("GET", url, proxy=proxy) as resp:
                        body = await resp.read()
                        data = json.loads(body.decode('utf-8'))
                        if data.get('args', {}).get('seed') != seed:
                            return False
        except CancelledError:
            raise
        except Exception:
            return False
        log.debug("Proxy %s is OK", addr)
        return True



# coding=utf-8


class UsageError(Exception):
    """
    CLI usage error.
    """

    def __init__(self, *args, **kwargs):
        self.print_help = kwargs.pop('print_help', True)
        super().__init__(*args, **kwargs)


class NetworkError(Exception):
    """
    Network error.
    """



# coding=utf-8

import re

from lxml import etree

ip_reg = re.compile(r'(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|\d{2,5})')


def is_ip(t):
    first = True
    for i in t.split('.'):
        if first:
            if int(i) == 0:
                return False
            first = False
        if int(i) > 255:
            return False
    return True


def is_port(t):
    t = int(t)
    # TODO HTTPS 443
    if t == 80:
        return True
    if 1024 < t < 65536:
        return True
    return False


def extract_proxies(html):
    try:
        root = etree.fromstring(html, parser=etree.HTMLParser())
        s = root.xpath('//body', smart_strings=False)
        text = ''
        for i in s:
            text += ' '.join(i.itertext())
    except Exception:
        return []
    res = []
    pre_ip = None
    for i in ip_reg.findall(text):
        if i.find('.') >= 0:
            pre_ip = i
        else:
            if pre_ip and is_ip(pre_ip) and is_port(i):
                res.append(pre_ip + ':' + i)
    return res



# coding=utf-8

proxy_pages = {
    '89ip': 'http://www.89ip.cn/tqdl.html?num=1000',
    '66ip': 'http://www.66ip.cn/mo.php?tqsl=1000',
    'kuaidaili': ['https://www.kuaidaili.com/free/inha/{}/'.format(i) for i in range(5, 0, -1)]
                 + ['https://www.kuaidaili.com/free/intr/{}/'.format(i) for i in range(5, 0, -1)],
    'data5u': ['http://www.data5u.com/free/gngn/index.shtml',
               'http://www.data5u.com/free/gnpt/index.shtml',
               'http://www.data5u.com/free/gwgn/index.shtml',
               'http://www.data5u.com/free/gwpt/index.shtml'],
    'crossin': ['http://lab.crossincode.com/proxy/', 'http://lab.crossincode.com/proxy/get/?num=20'],
    'ihuan': ['https://ip.ihuan.me/?page={}'.format(i) for i in range(50, 0, -1)]
}



# coding=utf-8

import copy
from collections import MutableMapping
import types

from freehp import defaultconfig


class BaseConfig(MutableMapping):
    def __init__(self, values=None):
        self.attributes = {}
        self.update(values)

    def __getitem__(self, opt_name):
        if opt_name not in self:
            return None
        return self.attributes[opt_name]

    def __contains__(self, name):
        return name in self.attributes

    def get(self, name, default=None):
        return self[name] if self[name] is not None else default

    def getbool(self, name, default=None):
        v = self.get(name, default)
        return getbool(v)

    def getint(self, name, default=None):
        v = self.get(name, default)
        return getint(v)

    def getfloat(self, name, default=None):
        v = self.get(name, default)
        return getfloat(v)

    def getlist(self, name, default=None):
        v = self.get(name, default)
        return getlist(v)

    def __setitem__(self, name, value):
        self.set(name, value)

    def set(self, name, value):
        self.attributes[name] = value

    def update(self, values):
        if values is not None:
            if isinstance(values, BaseConfig):
                for name in values:
                    self.set(name, values[name])
            else:
                for name, value in values.items():
                    self.set(name, value)

    def delete(self, name):
        del self.attributes[name]

    def __delitem__(self, name):
        del self.attributes[name]

    def copy(self):
        return copy.deepcopy(self)

    def __iter__(self):
        return iter(self.attributes)

    def __len__(self):
        return len(self.attributes)


def getbool(v):
    try:
        return bool(int(v))
    except (ValueError, TypeError):
        if v in ("True", "true"):
            return True
        if v in ("False", "false"):
            return False
    return None


def getint(v):
    try:
        return int(v)
    except (ValueError, TypeError):
        pass
    return None


def getfloat(v):
    try:
        return float(v)
    except (ValueError, TypeError):
        pass
    return None


def getlist(v):
    if v is None:
        return None
    if isinstance(v, str):
        v = v.split(",")
    elif not hasattr(v, "__iter__"):
        v = [v]
    return list(v)


class Config(BaseConfig):
    def __init__(self, values=None):
        super().__init__()
        for key in dir(defaultconfig):
            if not key.startswith("_"):
                value = getattr(defaultconfig, key)
                if not isinstance(value, (types.FunctionType, types.ModuleType, type)):
                    self.set(key.lower(), value)
        self.update(values)



# coding=utf-8

import aiohttp
from aiohttp import web
import async_timeout

from freehp.checker import HttpbinChecker


async def make_proxy_server(test_server, loop):
    async def process(request):
        async with aiohttp.ClientSession(loop=loop) as session:
            with async_timeout.timeout(60, loop=loop):
                async with session.request("GET", request.raw_path) as resp:
                    body = await resp.read()
                    return web.Response(status=resp.status,
                                        body=body,
                                        headers=resp.headers)

    app = web.Application()
    app.router.add_route("GET", "/{tail:.*}", process)
    server = await test_server(app)
    return server


class TestHttpbinChecker:
    async def test_check_proxy(self, test_server, loop):
        server = await make_proxy_server(test_server, loop)
        checker = HttpbinChecker(loop=loop)
        ok = await checker.check_proxy("{}:{}".format(server.host, server.port))
        assert ok is True
        ok = await checker.check_proxy("{}:{}".format(server.host, server.port - 1))
        assert ok is False



# coding=utf-8

import json
import time
import asyncio
import logging
from asyncio.queues import Queue
from collections import deque
import signal
from asyncio import CancelledError

from aiohttp import web

from freehp.spider import ProxySpider
from freehp.utils import load_object

log = logging.getLogger(__name__)


class ProxyManager:
    def __init__(self, config):
        self.config = config
        self.loop = asyncio.new_event_loop()

        self._checker = self._load_checker(config.get("checker_cls"))
        self._block_time = config.getint("block_time")
        self._proxy_queue = ProxyQueue(max_fail_times=config.getint("max_fail_times"))
        self._spider = ProxySpider.from_manager(self)
        self._spider.subscribe(self._add_proxy)

        self._proxy_db = {}

        self._futures = None
        self._wait_queue = Queue(loop=self.loop)
        self._is_running = False

    def start(self):
        if not self._is_running:
            self._is_running = True
            asyncio.set_event_loop(self.loop)
            self._futures = []
            self._init_server()
            self._init_checker()
            self._spider.open()
            self.loop.add_signal_handler(signal.SIGINT,
                                         lambda loop=self.loop: asyncio.ensure_future(self.shutdown(), loop=loop))
            self.loop.add_signal_handler(signal.SIGTERM,
                                         lambda loop=self.loop: asyncio.ensure_future(self.shutdown(), loop=loop))
            try:
                self.loop.run_forever()
            except Exception:
                log.error("Unexpected error occurred when run loop", exc_info=True)
                raise

    async def shutdown(self):
        if not self._is_running:
            return
        self._is_running = False
        log.info("Shutdown now")
        self._spider.close()
        if self._futures:
            for f in self._futures:
                f.cancel()
            self._futures = None
        await asyncio.sleep(0.001, loop=self.loop)
        self.loop.stop()

    def _init_server(self):
        bind = self.config.get('bind')
        log.info("Bind to '%s'", bind)
        app = web.Application(logger=log, loop=self.loop)
        app.router.add_route("GET", "/proxies", self.get_proxies)
        host, port = bind.split(":")
        port = int(port)
        self.loop.run_until_complete(
            self.loop.create_server(app.make_handler(access_log=None, loop=self.loop), host, port))

    async def _add_proxy(self, proxies):
        t = int(time.time())
        for p in proxies:
            try:
                proxy = self._proxy_db.get(p)
                if proxy and t - proxy.timestamp <= self._block_time:
                    continue
                proxy = ProxyInfo(p, t)
                self._proxy_db[p] = proxy
                await self._wait_queue.put(proxy)
            except CancelledError:
                raise
            except Exception:
                log.warning("Unexpected error occurred when add proxy '%s'", p, exc_info=True)

    def _init_checker(self):
        checker_clients = self.config.getint('checker_clients')
        log.info("Initialize checker, clients=%s", checker_clients)
        f = asyncio.ensure_future(self._find_expired_proxy_task(), loop=self.loop)
        self._futures.append(f)
        for i in range(checker_clients):
            f = asyncio.ensure_future(self._check_proxy_task(), loop=self.loop)
            self._futures.append(f)
        f = asyncio.ensure_future(self._remove_blocked_proxy_task(), loop=self.loop)
        self._futures.append(f)

    def _load_checker(self, cls_path):
        checker_cls = load_object(cls_path)
        if hasattr(checker_cls, "from_manager"):
            checker = checker_cls.from_manager(self)
        else:
            checker = checker_cls()
        return checker

    async def _find_expired_proxy_task(self):
        while True:
            proxy = self._proxy_queue.get_expired_proxy()
            if proxy is not None:
                await self._wait_queue.put(proxy)
            else:
                await asyncio.sleep(5, loop=self.loop)

    async def _check_proxy_task(self):
        check_interval = self.config.get('check_interval')
        while True:
            proxy = await self._wait_queue.get()
            ok = await self._checker.check_proxy(proxy.addr)
            t = int(time.time())
            proxy.timestamp = t + check_interval
            self._proxy_queue.feed_back(proxy, ok)

    async def _remove_blocked_proxy_task(self):
        while True:
            await asyncio.sleep(self._block_time, loop=self.loop)
            t = time.time()
            for i in list(self._proxy_db.keys()):
                if t - self._proxy_db[i].timestamp > self._block_time:
                    del self._proxy_db[i]

    async def get_proxies(self, request):
        params = request.rel_url.query
        count = params.get("count", 0)
        if count:
            count = int(count)
        detail = "detail" in params
        order = params.get('order', 'rate')
        log.info("GET '/', count=%s, detail=%s, order=%s", count, detail, order)
        proxy_list = self._get_proxies(count, detail=detail, order=order)
        return web.Response(body=json.dumps(proxy_list).encode("utf-8"),
                            charset="utf-8",
                            content_type="application/json")

    def _get_proxies(self, count, detail=False, order='rate'):
        t = self._proxy_queue.get_proxies()
        if count <= 0 or len(t) < count:
            count = len(t)
        if order == 'rate':
            t.sort(key=lambda k: k.rate, reverse=True)
        elif order == 'time':
            t.sort(key=lambda k: k.timestamp, reverse=True)
        t = t[:count]

        res = []
        for p in t:
            if detail:
                res.append({"addr": p.addr, "success": p.good, "fail": p.bad, "timestamp": p.timestamp})
            else:
                res.append(p.addr)
        return res


class ProxyQueue:
    def __init__(self, max_fail_times=2):
        self._max_fail_times = max_fail_times
        self._queue = deque()
        self._backup = deque()

    def get_proxies(self):
        res = [i for i in self._queue]
        return res

    def add_proxy(self, proxy):
        if proxy.fail == 0:
            self._queue.append(proxy)
        else:
            self._backup.append(proxy)

    def feed_back(self, proxy, ok):
        if ok:
            proxy.good += 1
            proxy.fail = 0
            self.add_proxy(proxy)
        else:
            proxy.bad += 1
            proxy.fail += 1
            if proxy.fail <= self._max_fail_times:
                self.add_proxy(proxy)

    def get_expired_proxy(self):
        t = int(time.time())
        p = None
        if len(self._queue) > 0 and len(self._backup) > 0:
            if t > self._queue[0].timestamp and t > self._backup[0].timestamp:
                if self._queue[0].timestamp > self._backup[0].timestamp:
                    p = self._backup.popleft()
                else:
                    p = self._queue.popleft()
            elif t > self._queue[0].timestamp:
                p = self._queue.popleft()
            elif t > self._backup[0].timestamp:
                p = self._backup.popleft()
        elif len(self._queue) > 0:
            if t > self._queue[0].timestamp:
                p = self._queue.popleft()
        elif len(self._backup) > 0:
            if t > self._backup[0].timestamp:
                p = self._backup.popleft()
        return p


class ProxyInfo:
    def __init__(self, addr, timestamp, *, good=0, bad=0, fail=1):
        self.addr = addr
        self.timestamp = timestamp
        self.good = good
        self.bad = bad
        self.fail = fail

    @property
    def rate(self):
        return self.good / (self.good + self.bad + 1.0)



# coding=utf-8

log_file = None
log_level = 'INFO'
log_format = '%(asctime)s %(name)s [%(levelname)s] %(message)s'
log_dateformat = '%Y-%m-%d %H:%M:%S'

daemon = False

bind = '0.0.0.0:6256'

block_time = 7200
max_fail_times = 2

check_interval = 300
checker_cls = 'freehp.checker.HttpbinChecker'
checker_clients = 100

scrap_interval = 300
spider_timeout = 30
spider_sleep_time = 5
spider_headers = {
    'Connection': 'keep-alive',
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
    'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/66.0.3359.181 Safari/537.36',
    'Accept-Encoding': 'gzip, deflate, sdch',
    'Accept-Language': 'zh-CN,zh;q=0.8'
}
proxy_pages = None



# coding=utf-8

import time
import asyncio
import logging
from asyncio import CancelledError

import aiohttp
import async_timeout

from freehp.extractor import extract_proxies

log = logging.getLogger(__name__)


class ProxySpider:
    def __init__(self, config, loop=None):
        self._proxy_pages = config.get('proxy_pages', {})
        log.debug('Details of proxy pages: %s', [i for i in self._proxy_pages])
        self._scrap_interval = config.getint("scrap_interval")
        self._timeout = config.getint("spider_timeout")
        self._sleep_time = config.getint("spider_sleep_time")
        self._headers = config.get("spider_headers", {})
        self._loop = loop or asyncio.get_event_loop()

        self._futures = None
        self._receivers = []

    @classmethod
    def from_manager(cls, manager):
        return cls(manager.config, loop=manager.loop)

    def subscribe(self, receiver):
        self._receivers.append(receiver)

    def open(self):
        self._futures = []
        for p in self._proxy_pages:
            f = asyncio.ensure_future(self._update_proxy_task(self._proxy_pages[p]), loop=self._loop)
            self._futures.append(f)

    def close(self):
        if self._futures:
            for f in self._futures:
                f.cancel()
            self._futures = None

    async def _update_proxy_task(self, urls):
        if not isinstance(urls, list):
            urls = [urls]
        while True:
            t = await self._update_proxy(urls)
            t = self._scrap_interval - t
            if t > self._sleep_time:
                await asyncio.sleep(t, loop=self._loop)

    async def _update_proxy(self, urls):
        start_time = time.time()
        for url in urls:
            retry_cnt = 3
            while retry_cnt > 0:
                retry_cnt -= 1
                try:
                    async with aiohttp.ClientSession(loop=self._loop) as session:
                        with async_timeout.timeout(self._timeout, loop=self._loop):
                            async with session.request("GET", url, headers=self._headers) as resp:
                                body = await resp.read()
                                body = body.decode('utf-8', errors='ignore')
                except CancelledError:
                    raise
                except Exception as e:
                    log.info("%s error occurred when scrap proxy on url=%s: %s", type(e), url, e)
                else:
                    retry_cnt = 0
                    proxies = extract_proxies(body)
                    log.debug("Find %s proxies on the page '%s'", len(proxies), url)
                    if proxies:
                        for r in self._receivers:
                            await r(proxies)
            await asyncio.sleep(self._sleep_time, loop=self._loop)
        return time.time() - start_time



# coding=utf-8

import os
from os.path import isfile

import logging
from importlib import import_module


def load_object(path):
    if isinstance(path, str):
        dot = path.rindex(".")
        module, name = path[:dot], path[dot + 1:]
        mod = import_module(module)
        return getattr(mod, name)
    return path


def load_config(fname):
    if fname is None or not isfile(fname):
        raise ValueError('{} is not a file'.format(fname))
    code = compile(open(fname, 'rb').read(), fname, 'exec')
    cfg = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": fname,
        "__doc__": None,
        "__package__": None
    }
    exec(code, cfg, cfg)
    return cfg


def configure_logging(name, config):
    log_level = config.get('log_level')
    log_format = config.get('log_format')
    log_dateformat = config.get('log_dateformat')
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    filename = config.get('log_file')
    if filename:
        handler = logging.FileHandler(filename)
    else:
        handler = logging.StreamHandler()
    handler.setLevel(log_level)
    formatter = logging.Formatter(log_format, log_dateformat)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def be_daemon():
    if os.fork():
        os._exit(0)
    os.setsid()
    if os.fork():
        os._exit(0)
    os.umask(0o22)
    os.closerange(0, 3)
    fd_null = os.open(os.devnull, os.O_RDWR)
    if fd_null != 0:
        os.dup2(fd_null, 0)
    os.dup2(fd_null, 1)
    os.dup2(fd_null, 2)



# coding=utf-8

import os
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

with open("README.rst") as fd:
    long_description = fd.read()


def read_version():
    p = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                     "freehp",
                     "version.py")
    with open(p, "rb") as f:
        return f.read().decode("utf-8").split("=")[-1].strip().strip('"')


class PyTest(TestCommand):
    user_options = []

    def run(self):
        import subprocess
        errno = subprocess.call([sys.executable, '-m', 'pytest', 'tests'])
        raise SystemExit(errno)


def main():
    if sys.version_info < (3, 5, 3):
        raise RuntimeError("Python 3.5.3+ is required")
    install_requires = [
        "aiohttp>=3.1.3,<4.0",
        "lxml>=4.1.0,<5.0"
    ]
    tests_requires = install_requires + ["pytest", "pytest-aiohttp"]
    setup(
        name="freehp",
        version=read_version(),
        url="https://github.com/jadbin/freehp",
        description="Providing Free HTTP proxies",
        long_description=long_description,
        author="jadbin",
        author_email="jadbin.com@hotmail.com",
        license="Apache 2",
        zip_safe=False,
        packages=find_packages(exclude=("tests",)),
        include_package_data=True,
        entry_points={
            "console_scripts": ["freehp = freehp.cli:main"]
        },
        install_requires=install_requires,
        tests_require=tests_requires,
        cmdclass={"test": PyTest},
        classifiers=[
            "License :: OSI Approved :: Apache Software License",
            "Intended Audience :: Developers",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Topic :: Internet :: WWW/HTTP"
        ]
    )


if __name__ == "__main__":
    main()



# coding=utf-8

import logging
from os.path import abspath, isfile

from freehp.errors import UsageError
from freehp import utils
from freehp.version import __version__
from freehp.manager import ProxyManager
from freehp import defaultconfig
from freehp import config
from freehp import squid

log = logging.getLogger(__name__)


class Command:
    def __init__(self):
        self.config = config.Config()
        self.exitcode = 0

    @property
    def name(self):
        return ""

    @property
    def syntax(self):
        return ""

    @property
    def short_desc(self):
        return ""

    @property
    def long_desc(self):
        return self.short_desc

    def add_arguments(self, parser):
        pass

    def process_arguments(self, args):
        pass

    def run(self, args):
        raise NotImplementedError


class RunCommand(Command):
    @property
    def name(self):
        return "run"

    @property
    def syntax(self):
        return "[OPTIONS]"

    @property
    def short_desc(self):
        return "Run spider to scrap free HTTP proxies"

    def add_arguments(self, parser):
        parser.add_argument('-c', '--config', metavar='FILE', help='configuration file')
        parser.add_argument('-b', '--bind', dest='bind', metavar='ADDRESS', default=defaultconfig.bind,
                            help='the socket to bind')
        parser.add_argument('-d', '--daemon', dest='daemon', action='store_true', default=defaultconfig.daemon,
                            help='run in daemon mode')
        parser.add_argument("-l", "--log-level", dest="log_level", metavar="LEVEL", default=defaultconfig.log_level,
                            help="log level")
        parser.add_argument("--log-file", dest="log_file", metavar="FILE", default=defaultconfig.log_file,
                            help="log file")
        parser.add_argument("-s", "--set", dest="set", action="append", default=[], metavar="NAME=VALUE",
                            help="set/override setting (may be repeated)")

    def process_arguments(self, args):
        if args.config:
            if isfile(args.config):
                for k, v in utils.load_config(args.config).items():
                    self.config.set(k, v)
            else:
                self.exitcode = 1
                print("Error: Cannot find '{}'".format(abspath(args.config)))
                return
        if args.bind:
            self.config.set('bind', args.bind)
        if args.daemon:
            self.config.set('daemon', True)
        if args.log_level:
            self.config.set("log_level", args.log_level)
        if args.log_file:
            self.config.set('log_file', args.log_file)
        try:
            self.config.update(dict(x.split("=", 1) for x in args.set))
        except ValueError:
            raise UsageError("Invalid -s value, use -s NAME=VALUE", print_help=False)

    def run(self, args):
        if self.config.getbool('daemon'):
            utils.be_daemon()
        utils.configure_logging('freehp', self.config)
        agent = ProxyManager(self.config)
        agent.start()


class SquidCommand(Command):
    @property
    def name(self):
        return "squid"

    @property
    def syntax(self):
        return "[OPTIONS] <DEST_FILE>"

    @property
    def short_desc(self):
        return "Append proxies to the configuration of squid"

    def add_arguments(self, parser):
        parser.add_argument('dest_file', metavar='DEST_FILE', nargs=1,
                            help='where the squid configuration file is')
        parser.add_argument('-a', '--address', dest='address', metavar='ADDRESS', default=squid.DEFAULT_FREEHP_ADDRESS,
                            help='the address of freehp')
        parser.add_argument('--template', dest='template', metavar='FILE',
                            help='the template of squid configuration, default is <DEST_FILE>')
        parser.add_argument('--squid', dest='squid', metavar='squid', default=squid.DEFAULT_SQUID,
                            help='the name of squid command')
        parser.add_argument('-d', '--daemon', dest='daemon', action='store_true', default=False,
                            help='run in daemon mode')
        parser.add_argument('--max-num', dest='max_num', type=int, metavar='NUM', default=squid.DEFAULT_MAX_NUM,
                            help='max number of proxies, 0 for unlimited')
        parser.add_argument('--update-interval', dest='update_interval', type=int, metavar='SECONDS',
                            default=squid.DEFAULT_UPDATE_INTERVAL, help='update interval in seconds')
        parser.add_argument('--timeout', dest='timeout', type=int, metavar='SECONDS', default=squid.DEFAULT_TIMEOUT,
                            help='timeout in seconds')
        parser.add_argument('--once', dest='once', action='store_true', default=False,
                            help='run only once')
        parser.add_argument("-l", "--log-level", dest="log_level", metavar="LEVEL", default=defaultconfig.log_level,
                            help="log level")
        parser.add_argument("--log-file", dest="log_file", metavar="FILE", default=defaultconfig.log_file,
                            help="log file")

    def process_arguments(self, args):
        args.dest_file = args.dest_file[0]
        if not args.template:
            if not isfile(args.dest_file):
                raise UsageError('The template of squid configuration is not specified', print_help=False)
            args.template = args.dest_file
        if args.log_level:
            self.config.set("log_level", args.log_level)
        if args.log_file:
            self.config.set('log_file', args.log_file)

    def run(self, args):
        if config.getbool(args.daemon):
            utils.be_daemon()
        utils.configure_logging('freehp', self.config)
        s = squid.Squid(args.dest_file, args.template, freehp_address=args.address, max_num=args.max_num,
                        update_interval=args.update_interval, timeout=args.timeout, once=args.once)
        s.start()


class VersionCommand(Command):
    @property
    def name(self):
        return "version"

    @property
    def short_desc(self):
        return "Print the version"

    def run(self, args):
        print("freehp version {0}".format(__version__))
